import java.io.*;
import java.util.StringTokenizer;

/** This class does the job of aligning a given input RNA sequence with 
 * the csHMM representing miRNA secondary structure, in the process 
 * generating the most likely secondary structure for the
 * given linear sequence, assuming it to be a miRNA-precursor. The program 
 * takes a file containing the csHMM parameters (this should be in the 
 * format generated by Learn.java) as a command line argument, and then 
 * inputs a set of miRNA sequences and generates the corresponding 
 * secondary structures. All input and output is from standard I/O.
 *
 * @author Sumeet Agarwal
 * @version 7.3.2010.
 */ 

public class Align2
{
    int M = 8;
    int S = 5;
    int MAX_LEN = 500;  
    double neg = -1;	 
    double gamma[][][][] = new double[MAX_LEN][MAX_LEN][M][M];   
    LV lambda[][][][] = new LV[MAX_LEN][MAX_LEN][M][M];  
    double exp[][] = new double[S][4];
    double cexp[][][] = new double[2][4][4];
    double trans[][] = new double[S][M];		 
    double ctrans[][][] = new double[2][5][M];
    static InputStreamReader reader = new InputStreamReader(System.in);
    static BufferedReader console = new BufferedReader(reader);

    /* Default constructor */	    
    public Align2(){
    }		
    
    /* The main function; just reads a set of RNA sequences as input and outputs corresponding
     * secondary structures. */
    public static void main(String args[]) throws IOException 
    {
        Align2 a = new Align2();
	try{		
		a.getParams(args[0]);	
	}
	catch(Exception e){
		System.out.print("Please enter parameter file for csHMM: ");	
		a.getParams(console.readLine());
	}
        String in;
        while(true){
	    try{
                in = console.readLine();
                if(in == null)  
                    break;
		String[] ret = a.genPath(in);
		double p = (Double.valueOf(ret[0])).doubleValue();
		String pi = ret[1];
                //System.out.println("Most likely secondary structure is:\n");
		String[] l = a.genFigure(in, pi);
		for(int i=0; i<5; i++)
			System.out.println(l[i]);
		System.out.println(); 
                /*System.out.println("Feature values:");
		FV f = a.getFeatures(in, p, pi);
		f.display();*/
            }
            catch(Exception e){
        	    System.out.println("Exception: "+e);
	    }	
        }
    }
    
    /* The alignment function; implements the algorithm to generate the most likely sequence and
     * its likelihood. 
     * 
     * @param in The input string to be aligned
     * @param i The starting position in the string
     * @param j The ending position in the string
     * @param v The starting state for the csHMM
     * @param w The ending state for the csHMM
     *
     * @return The 'gamma' value; representing the likelihood of the alignment generated		     */	
    public double gamma(String in, int i, int j, int v, int w){
        if(gamma[i][j][v][w] == neg){
            if(i==j && v==w){
                int xi = toNum(in, i);          
                if(v==1 || v==5 || v==6);
                else
                    gamma[i][j][v][w] = exp[v][xi];
                lambda[i][j][v][w] = new LV(0,0,0,0);    
            }
            else if(i>=j)
                lambda[i][j][v][w] = new LV(0,0,0,0); 
            else{
                int xi = toNum(in, i);
                int xj = toNum(in, j);
                if(v==5 || v==6 || w==1)
                    lambda[i][j][v][w] = new LV(0,0,0,0); 
                else if(v==1 && (w==5 || w==6)){
                    if(j==i+1){
                        gamma[i][j][v][w] = exp[v][xi]*trans[v][w]*cexp[w%5][xi][xj];    
                        lambda[i][j][v][w] = new LV(0,0,0,0);
                    }
                    else{
                        double num=0;
                        int u1s=0, u2s=0;
                        for(int u1=1; u1<M-1; u1++){
                            for(int u2=1; u2<M-1; u2++){
                                double t;
                                if(u2 >= 5)
                                    t = ctrans[u2%5][xi][w];                              
                                else
                                    t = trans[u2][w];
                                num = exp[v][xi]*trans[v][u1]*t*cexp[w%5][xi][xj]*gamma(in, i+1, j-1, u1, u2);
                                if(num > gamma[i][j][v][w]){
                                    gamma[i][j][v][w] = num;
                                    u1s = u1;
                                    u2s = u2;
                                }
                            }
                        }
                        lambda[i][j][v][w] = new LV(i+1, j-1, u1s, u2s);
                    }
                }       
                else if(v==1){
                    double num=0;
                    int us=0;
                    for(int u=1; u<M-1; u++){
                        double t;
                        if(u >= 5)
                            t = ctrans[u%5][xi][w];
                        else
                            t = trans[u][w];
                        num = gamma(in,i,j-1,v,u)*t*exp[w][xj]; 
                        if(num > gamma[i][j][v][w]){
                            gamma[i][j][v][w] = num;
                            us = u;
                        }
                    }
                    lambda[i][j][v][w] = new LV(i, j-1, v, us);
                }
                else{   
                    double num=0;
                    int us=0;
                    for(int u=1; u<M-1; u++){
                        num = exp[v][xi]*trans[v][u]*gamma(in,i+1,j,u,w); 
                        if(num > gamma[i][j][v][w]){                                                                  
	          gamma[i][j][v][w] = num;
                            us = u;
                        }
                    }
                    lambda[i][j][v][w] = new LV(i+1, j, us, w);
                }
            }
        }
        return gamma[i][j][v][w];       
    }

    /* Converts the bases in the input sequence to corresponding numbers.
     *
     * @param in The input string
     * @param i The position in the input string to be converted to a number
     *
     * @return The integer value corresponding to the base at the given position
     */ 	            
    public int toNum(String in, int i){
        char ci = in.charAt(i);
        int xi = -1;
        switch(ci){
            case 'a':
            case 'A':	
                xi = 0;
                break;
            case 'c':
            case 'C':	
                xi = 1;
                break;
            case 'g':
            case 'G':
                xi = 2;
                break;
            case 'u':
            case 'U':
            case 't':
            case 'T':	
                xi = 3;
                break;
        }
        return xi;
    }

    /* Uses the gamma() function to generate the final most likely path for a given input string.
     *
     * @param in The input string to be aligned
     *
     * @return An array having two elements: 1. The likelihood of the alignment generated; 2. The
     * corresponding state sequence
     */	
    public String[] genPath(String in){	
	double p=neg;
        for(int i=0; i<MAX_LEN; i++)
        	for(int j=0; j<MAX_LEN; j++)
                        for(int v=0; v<M; v++)
                		for(int w=0; w<M; w++)
                                	gamma[i][j][v][w] = neg;
        int vs=0, ws=0;
	for(int v=1; v<M-1; v++){   
        	for(int w=1; w<M-1; w++){
                	double t;
                    	if(w>=5)
                        	t = ctrans[w%5][4][M-1];
                    	else
                        	t = trans[w][M-1];
                    	double num = trans[0][v]*gamma(in, 0, in.length()-1, v, w)*t;
                    	if(num > p){
                       		p = num;
				vs = v;
                        	ws = w;
                    	}
                }
	}
	String[] ret = new String[2];
	ret[0] = Double.toString(p);
        LV l = new LV(0, in.length()-1, vs, ws);
        Stack tl = new Stack(MAX_LEN);
        Stack tr = new Stack(MAX_LEN); 
        while(!l.toString().equals("0000")){
            int i = l.i;
            int j = l.j;         
            int v = l.v;
            int w = l.w;
            l = lambda[i][j][v][w];     
            if(l.i != i) 
                tl.push(v);
            if(l.j != j)
                tr.push(w);
        }
        if(!tl.empty())
            tl.pop();
        while(!tl.empty()){
            int v = tl.pop();
            tr.push(v);
        }
        String pi = "";
        while(!tr.empty()){
            int v = tr.pop();
            pi += v;
        }
	ret[1] = pi;	
        return ret;
    }   
    
    /* Generates the secondary structure as a graphical display.
     *
     * @param seq The input RNA sequence
     * @param path A corresponding state sequence from the csHMM
     *
     * @return A string array, consisting of 5 lines to be printed in sequence to get the figure
     */
    public String[] genFigure(String seq, String path)
    {
	String[] l = new String[5];
	for(int i=0; i<5; i++)
		l[i] = "";
	int left = 0;
	int right = path.length()-1;
	while(left <= right){
		if(path.charAt(left) == '1'){
			while(path.charAt(right) == '4'){
				l[0] += "-";
				l[1] += " ";
				l[2] += " ";
				l[3] += " ";
				l[4] += seq.charAt(right);	
				right--;
			}
			if(path.charAt(right) == '5'){
				l[0] += " ";
				l[1] += seq.charAt(left);
				l[2] += "|";
				l[3] += seq.charAt(right);
				l[4] += " ";
				right--;
			}
			else{
				l[0] += seq.charAt(left);
				l[1] += " ";
				l[2] += " ";
				l[3] += " ";
				l[4] += seq.charAt(right);
				right--;
			}
			left++;
		}
		else if(path.charAt(left) == '2'){
			l[0] += seq.charAt(left);
			l[1] += " ";
			l[2] += " ";
			l[3] += " ";
			l[4] += "-";
			left++;
		}
		else{
			if(left == right)
				l[2] += seq.charAt(left);
			else if(right-left <= 2){
				l[1] += seq.charAt(left);
				l[3] += seq.charAt(right);
			}
			else{
				l[0] += seq.charAt(left);
				l[1] += " ";
				l[2] += " ";
				l[3] += " ";
				l[4] += seq.charAt(right);
			}
			left++;
			right--;
		}
	}
	return l;
    }

    /* Gets the set of secondary structure feature values for a given alignment.
     * @param seq The input RNA sequence
     * @param prob The likelihood of the generated alignment
     * @param path The state sequence for the generated alignment
     *
     * @return The feature vector for the sequence
     */	
    public FV getFeatures(String seq, double prob, String path){
	FV f = new FV(seq, prob, path);
	return f;		
    }

    /* Gets model parameters from a file.
     *
     * @param file The name of the parameter file
     */			
    public void getParams(String file) throws IOException{
	FileInputStream str = new FileInputStream(file);
        InputStreamReader isr = new InputStreamReader(str);
        BufferedReader buf = new BufferedReader(isr);       
        for(int i=0; i<S; i++){
                StringTokenizer t = new StringTokenizer(buf.readLine());
                for(int j=0; j<M; j++)
                    trans[i][j] = Double.parseDouble(t.nextToken());
        }
        buf.readLine();
        for(int i=0; i<S; i++){
                StringTokenizer t = new StringTokenizer(buf.readLine());
                for(int j=0; j<4; j++)
                    exp[i][j] = Double.parseDouble(t.nextToken());
        }
        buf.readLine();
        for(int h=0; h<2; h++){
                for(int i=0; i<5; i++){
                    StringTokenizer t = new StringTokenizer(buf.readLine());
                    for(int j=0; j<M; j++)
                        ctrans[h][i][j] = Double.parseDouble(t.nextToken());
                }
                buf.readLine();
        }
        for(int h=0; h<2; h++){
                for(int i=0; i<4; i++){
                    StringTokenizer t = new StringTokenizer(buf.readLine());
                    for(int j=0; j<4; j++)
                        cexp[h][i][j] = Double.parseDouble(t.nextToken());
                }
                buf.readLine();
        }
	str.close();
    }		
}	
